subroutine PerpendicularEfield(E, B, Eperp)

  ! No implicit variables
  implicit none
  
  double precision B(3)
  double precision E(3), Eperp(3)
  double precision u(3)

  ! Unit vector in the B direction
  u = B / sqrt( B(1)**2.0d0 + B(2)**2.0d0 + B(3)**2.0d0 )

  ! The component of E which is perpendicular to B
  Eperp = E - u*sum(u*E)

end subroutine PerpendicularEfield

!------------------------------------------------------------------------

subroutine Boris(velocity_in,velocity_out,E,B,dt,mass,charge)


! This is the velocity update. It uses the Boris algorithm as described 
! by, e.g. Birdsall and Langdon.
!
!@Book{BL91,
!  author =	 {C. K. Birdsall and A. B. Langdon},
!  title =	 {Plasma physics via computer simulation},
!  publisher =	 "IOP Publishing Ltd",
!  address =	 "Dirac House, Temple Back, Bristol BS1 6BE, UK",
!  year =	 1991,
!  note =	 "{ISBN} 0 7503 1025 1"
!}
!
!PAGES:61-63
! No implicit variables
  implicit none

! Parameters
  double precision velocity_in(3), velocity_out(3)
  double precision E(3), B(3)
  double precision dt, mass, charge


! Local variables
  double precision s(3), t(3), vplus(3), vmin(3), &
       vmid(3), eta, etadt2
  integer i


! Set constant local variables
  eta = charge/mass
  etadt2 = eta*dt/2.0d0

  
! Calculate t and vmin (first half acceleration)
  do i = 1, 3
     t(i) = etadt2*dble(B(i))
     vmin(i) = velocity_in(i)+etadt2*E(i)
  end do

! Cross product (vmid = vmin + vmin x t)
  vmid(1) = vmin(1) + vmin(2)*t(3)-vmin(3)*t(2)
  vmid(2) = vmin(2) + vmin(3)*t(1)-vmin(1)*t(3)
  vmid(3) = vmin(3) + vmin(1)*t(2)-vmin(2)*t(1)

! Calculate s
  do i = 1, 3
     s(i) = 2.0d0*t(i)/(1.0d0 + t(1)**2.0d0 + t(2)**2.0d0 + t(3)**2.0d0)
  end do

! Another cross product (vplus = vmin + vmid x s)
  vplus(1) = vmin(1) + vmid(2)*s(3)-vmid(3)*s(2)
  vplus(2) = vmin(2) + vmid(3)*s(1)-vmid(1)*s(3)
  vplus(3) = vmin(3) + vmid(1)*s(2)-vmid(2)*s(1)

! Loop and set new velocities (second half acceleration)
  do i= 1, 3
     velocity_out(i) = vplus(i) + etadt2*E(i)
  end do

  return
end subroutine Boris

!------------------------------------------------------------------------

subroutine DustyBoris(velocity_in,velocity_out,E,B,dt,mass,charge,Flight)


! This is the velocity update. It uses the Boris algorithm as described 
! by, e.g. Birdsall and Langdon.
!
!@Book{BL91,
!  author =	 {C. K. Birdsall and A. B. Langdon},
!  title =	 {Plasma physics via computer simulation},
!  publisher =	 "IOP Publishing Ltd",
!  address =	 "Dirac House, Temple Back, Bristol BS1 6BE, UK",
!  year =	 1991,
!  note =	 "{ISBN} 0 7503 1025 1"
!}
!PAGES:61-63
!
! A force corresponding to the light pressure is added in the negative 
! x direction.

! No implicit variables
  implicit none

! Parameters
  double precision velocity_in(3), velocity_out(3)
  double precision E(3), B(3)
  double precision dt, mass, charge, Flight


! Local variables
  double precision s(3), t(3), vplus(3), vmin(3), &
       vmid(3), eta, etadt2, dvLight
  integer i


! Set constant local variables
  eta = charge/mass
  etadt2 = eta*dt/2.0d0
  dvLight = -Flight/mass*dt/2.0d0
  
! Calculate t and vmin (first half acceleration)
  do i = 1, 3
     t(i) = etadt2*dble(B(i))
     vmin(i) = velocity_in(i)+etadt2*E(i)
  end do
  vmin(1) = vmin(1) + dvLight

! Cross product (vmid = vmin + vmin x t)
  vmid(1) = vmin(1) + vmin(2)*t(3)-vmin(3)*t(2)
  vmid(2) = vmin(2) + vmin(3)*t(1)-vmin(1)*t(3)
  vmid(3) = vmin(3) + vmin(1)*t(2)-vmin(2)*t(1)

! Calculate s
  do i = 1, 3
     s(i) = 2.0d0*t(i)/(1.0d0 + t(1)**2.0d0 + t(2)**2.0d0 + t(3)**2.0d0)
  end do

! Another cross product (vplus = vmin + vmid x s)
  vplus(1) = vmin(1) + vmid(2)*s(3)-vmid(3)*s(2)
  vplus(2) = vmin(2) + vmid(3)*s(1)-vmid(1)*s(3)
  vplus(3) = vmin(3) + vmid(1)*s(2)-vmid(2)*s(1)

! Loop and set new velocities (second half acceleration)
  do i= 1, 3
     velocity_out(i) = vplus(i) + etadt2*E(i)
  end do
  velocity_out(1) = velocity_out(1) + dvLight

  return
end subroutine DustyBoris

!------------------------------------------------------------------------

subroutine SpeedMomentAngle(vv,mu,pangle,velocity,B,mass)

! No implicit variables
  implicit none
  double precision mass
  double precision velocity(3), vv, vv2, mu, pangle

! local variables
  double precision bb, bb2, vperp2, B(3)

  bb2 = sum(B**2.0d0)   ! magnitude of B squared
  bb  = sqrt(bb2)       ! magnitude of B

  vv2= sum(velocity**2.0d0)
  vv = sqrt(vv2)

  !vperp2 = |v|^2 - (vB)^2/|B|^2
  ! mu = mass*vperp2/(2*|B|)
  mu = 0.5d0*mass*( vv2 - (sum(velocity*B)**2.0d0)/bb2 )/bb

  !angle = acos (vB/|v||B|) -> direction of vB
  pangle=acos(sum(velocity*B)/(vv*bb))

  return
  
end subroutine SpeedMomentAngle

!------------------------------------------------------------------------

subroutine InterpolateVectorField(E, Nx, Ny, Nz, dxyz, xmin, ymin, zmin, &
     x0, y0, z0, E0)

  implicit none

  integer Nx, Ny, Nz
  double precision E(3,Nx,Ny,Nz), dxyz(3), xmin, ymin, zmin, &
       x0, y0, z0, E0(3)

  integer Ng(3), ff, ii, jj, kk
  double precision rest(3), drest(3), FieldE(2,2,2,3), F(8), dxdydz

  dxdydz = dxyz(1)*dxyz(2)*dxyz(3)

  Ng(1)    = int(dint((x0-xmin)/dxyz(1))) + 1
  rest(1)  = dmod((x0-xmin),dxyz(1))
  drest(1) = dxyz(1)-rest(1)
  Ng(2)    = int(dint((y0-ymin)/dxyz(2))) + 1
  rest(2)  = dmod((y0-ymin),dxyz(2))
  drest(2) = dxyz(2)-rest(2)
  Ng(3)    = int(dint((z0-zmin)/dxyz(3))) + 1
  rest(3)  = dmod((z0-zmin),dxyz(3))
  drest(3) = dxyz(3)-rest(3)

! If the grid is specified too small it is possible that the particle 
! position is outside of the grid. If that happens we set E0=0 and return.
  if ( (minval(Ng)<=0) .or. (Ng(1)>=Nx) .or. (Ng(2)>=Ny) .or. (Ng(3)>=Nz) ) then
     E0 = 0.0d0
     return
  end if

! Find fields Fx, Fy, Fz  at eight closest gridpoints
  do ff = 1, 3
     do kk = 0, 1
        do jj = 0, 1
           do ii = 0, 1
              FieldE(ii+1,jj+1,kk+1,ff) = E(ff,Ng(1)+ii,Ng(2)+jj,Ng(3)+kk)
           end do
        end do
     end do
  end do

  F(1) = drest(1) * drest(2) * drest(3)
  F(2) =  rest(1) * drest(2) * drest(3)
  F(3) = drest(1) *  rest(2) * drest(3)
  F(4) = drest(1) * drest(2) *  rest(3)
  F(5) =  rest(1) *  rest(2) * drest(3)
  F(6) = drest(1) *  rest(2) *  rest(3)  
  F(7) =  rest(1) * drest(2) *  rest(3)
  F(8) =  rest(1) *  rest(2) *  rest(3)

  do ff = 1, 3
     E0(ff) = FieldE(1,1,1,ff)*F(1) + FieldE(2,1,1,ff)*F(2) + &
          FieldE(1,2,1,ff)*F(3) + FieldE(1,1,2,ff)*F(4) + &
          FieldE(2,2,1,ff)*F(5) + FieldE(1,2,2,ff)*F(6) + &
          FieldE(2,1,2,ff)*F(7) + FieldE(2,2,2,ff)*F(8)
       
     E0(ff) = E0(ff)/dxdydz

  end do

end subroutine InterpolateVectorField

!------------------------------------------------------------------------

subroutine FindFields(Ngridlevels, grid, x0, y0, z0, E0, B0)

  use SpecificTypes

  implicit none

  integer Ngridlevels
  double precision x0, y0, z0, E0(3), B0(3)
  type (GridLevel) grid(Ngridlevels)

  integer Ng(3), ff, ii, jj, kk, gl
  double precision rest(3), drest(3), FieldE(2,2,2,3), FieldB(2,2,2,3), &
       F(8), dxdydz

! Initialise output fields
  E0 = 0.0d0
  B0 = 0.0d0
  

! First find which level we are on.
  do ii = 1, Ngridlevels
     gl = ii
     if ( x0<grid(ii)%xmin .or. x0>grid(ii)%xmax .or. &
          y0<grid(ii)%ymin .or. y0>grid(ii)%ymax .or. &
          z0<grid(ii)%zmin .or. z0>grid(ii)%zmax ) then 
        gl = ii-1
        if ( gl .eq. 0 ) then
           return
        end if
        exit
     end if
  end do

  dxdydz = grid(gl)%dxyz(1)*grid(gl)%dxyz(2)*grid(gl)%dxyz(3)

  Ng(1)    = int(dint((x0-grid(gl)%xmin)/grid(gl)%dxyz(1))) + 1
  rest(1)  = dmod((x0-grid(gl)%xmin),grid(gl)%dxyz(1))
  drest(1) = grid(gl)%dxyz(1)-rest(1)
  Ng(2)    = int(dint((y0-grid(gl)%ymin)/grid(gl)%dxyz(2))) + 1
  rest(2)  = dmod((y0-grid(gl)%ymin),grid(gl)%dxyz(2))
  drest(2) = grid(gl)%dxyz(2)-rest(2)
  Ng(3)    = int(dint((z0-grid(gl)%zmin)/grid(gl)%dxyz(3))) + 1
  rest(3)  = dmod((z0-grid(gl)%zmin),grid(gl)%dxyz(3))
  drest(3) = grid(gl)%dxyz(3)-rest(3)


! Find fields Fx, Fy, Fz  at eight closest gridpoints
  do ff = 1, 3
     do kk = 0, 1
        do jj = 0, 1
           do ii = 0, 1
              FieldE(ii+1,jj+1,kk+1,ff) = &
                   grid(gl)%E(ff,Ng(1)+ii,Ng(2)+jj,Ng(3)+kk)
              FieldB(ii+1,jj+1,kk+1,ff) = &
                   grid(gl)%B(ff,Ng(1)+ii,Ng(2)+jj,Ng(3)+kk)
           end do
        end do
     end do
  end do

  F(1) = drest(1) * drest(2) * drest(3)
  F(2) =  rest(1) * drest(2) * drest(3)
  F(3) = drest(1) *  rest(2) * drest(3)
  F(4) = drest(1) * drest(2) *  rest(3)
  F(5) =  rest(1) *  rest(2) * drest(3)
  F(6) = drest(1) *  rest(2) *  rest(3)  
  F(7) =  rest(1) * drest(2) *  rest(3)
  F(8) =  rest(1) *  rest(2) *  rest(3)

  do ff = 1, 3
     E0(ff) = FieldE(1,1,1,ff)*F(1) + FieldE(2,1,1,ff)*F(2) + &
          FieldE(1,2,1,ff)*F(3) + FieldE(1,1,2,ff)*F(4) + &
          FieldE(2,2,1,ff)*F(5) + FieldE(1,2,2,ff)*F(6) + &
          FieldE(2,1,2,ff)*F(7) + FieldE(2,2,2,ff)*F(8)
     E0(ff) = E0(ff)/dxdydz
     B0(ff) = FieldB(1,1,1,ff)*F(1) + FieldB(2,1,1,ff)*F(2) + &
          FieldB(1,2,1,ff)*F(3) + FieldB(1,1,2,ff)*F(4) + &
          FieldB(2,2,1,ff)*F(5) + FieldB(1,2,2,ff)*F(6) + &
          FieldB(2,1,2,ff)*F(7) + FieldB(2,2,2,ff)*F(8)
     B0(ff) = B0(ff)/dxdydz

  end do

end subroutine FindFields

!------------------------------------------------------------------------

subroutine  FindFluct(rin, rout, Ax, Ay, Az, ffluct, x, y, z, &
             fluctphase, t, E1)

  implicit none

  double precision rin, rout, Ax, Ay, Az, ffluct, x, y, z
  double precision fluctphase, t, E1(3)

  double precision rr, pi, geometryfactor, phasefactor

  parameter(pi = 3.14159265358979323846264338327950288419716d0)

  E1 = 0.0d0 ! Initialise

  rr = sqrt(x**2.0d0 + y**2.0d0 + z**2.0d0)

  if ( (rr > rin) .and. (rr < rout) ) then
     geometryfactor = cos(pi * (rr-0.5d0*(rout+rin))/(rout-rin))
     phasefactor = sin(2*pi*ffluct*t + fluctphase)
     E1(1) = Ax * geometryfactor * phasefactor
     E1(2) = Ay * geometryfactor * phasefactor
     E1(3) = Az * geometryfactor * phasefactor
  end if

end subroutine FindFluct

!------------------------------------------------------------------------

subroutine AccumulateDensity(Nx, Ny, Nz, dens, dxyz, xmin, ymin, zmin, &
     Nsteps, position, Nend, dt)

  implicit none

  integer Nx, Ny, Nz, Nsteps, Nend
  double precision dens(Nx,Ny,Nz), dxyz(3), xmin, ymin, zmin
  double precision position(3,Nsteps), dt

  integer tt, Ng(3)
  double precision dV, rest(3), drest(3), F1, F2, F3, F4, F5, F6, F7, F8
  double precision lambda

  dV = dxyz(1)*dxyz(2)*dxyz(3)
  lambda = dt/dV

  do tt = 1, Nend

     ! Find out where on the grid the particle is
     Ng(1)    = int((position(1,tt)-xmin)/dxyz(1)) + 1
     rest(1)  = mod((position(1,tt)-xmin),dxyz(1)) 
     drest(1) = dxyz(1)-rest(1)
     Ng(2)    = int((position(2,tt)-ymin)/dxyz(2)) + 1 
     rest(2)  = mod((position(2,tt)-ymin),dxyz(2)) 
     drest(2) = dxyz(2)-rest(2)
     Ng(3)    = int((position(3,tt)-zmin)/dxyz(3)) + 1
     rest(3)  = mod((position(3,tt)-zmin),dxyz(3)) 
     drest(3) = dxyz(3)-rest(3)

     ! Start interpolation
     F1 = drest(1)*drest(2)*drest(3)/dV
     F2 = rest(1)*drest(2)*drest(3)/dV
     F3 = drest(1)*rest(2)*drest(3)/dV
     F5 = rest(1)*rest(2)*drest(3)/dV
     F4 = drest(1)*drest(2)*rest(3)/dV
     F7 = rest(1)*drest(2)*rest(3)/dV
     F6 = drest(1)*rest(2)*rest(3)/dV
     F8 = rest(1)*rest(2)*rest(3)/dV

     dens(Ng(1),Ng(2),Ng(3)) = dens(Ng(1),Ng(2),Ng(3)) + lambda*F1
     dens(Ng(1)+1,Ng(2),Ng(3)) = dens(Ng(1)+1,Ng(2),Ng(3)) + lambda*F2
     dens(Ng(1),Ng(2)+1,Ng(3)) = dens(Ng(1),Ng(2)+1,Ng(3)) + lambda*F3
     dens(Ng(1)+1,Ng(2)+1,Ng(3)) = dens(Ng(1)+1,Ng(2)+1,Ng(3)) + lambda*F5
     dens(Ng(1),Ng(2),Ng(3)+1) = dens(Ng(1),Ng(2),Ng(3)+1) + lambda*F4
     dens(Ng(1)+1,Ng(2),Ng(3)+1) = dens(Ng(1)+1,Ng(2),Ng(3)+1) + lambda*F7
     dens(Ng(1),Ng(2)+1,Ng(3)+1) = dens(Ng(1),Ng(2)+1,Ng(3)+1) + lambda*F6
     dens(Ng(1)+1,Ng(2)+1,Ng(3)+1)= dens(Ng(1)+1,Ng(2)+1,Ng(3)+1) + lambda*F8

  end do

end subroutine AccumulateDensity

!------------------------------------------------------------------------

subroutine AccumulatefProbe(Nprobes, probe, Nsteps, pos, vel, Nend, dt)

  use SpecificTypes

  implicit none

  integer Nprobes, Nsteps, Nend
  type(fprobe) probe(Nprobes)
  double precision pos(3,Nsteps), vel(3,Nsteps), dt

  integer ii, jj, tt, Ng(3)
  double precision dV, dVv, rest(3), drest(3), F1, F2, F3, F4, F5, F6, F7, F8
  double precision lambda, pi, probepos(3), rr

  parameter(pi = 3.14159265358979323846264338327950288419716d0)

  do ii = 1, Nprobes
     probepos = (/ probe(ii)%xc, probe(ii)%yc, probe(ii)%zc /)
     dV = 4.0d0*pi*probe(ii)%rprobe**3.0d0 / 3.0d0
     dVv = probe(ii)%dvx * probe(ii)%dvy * probe(ii)%dvz
     lambda = dt / (dV*dVv)

     do tt = 1, Nend
        ! Check if the particle is in the probe volume
        rr = sqrt(sum((pos(:,tt)-probepos)**2.0d0))
        if ( rr <= probe(ii)%rprobe ) then
           ! Find out where on the velocity grid the particle is
           Ng(1)    = int((vel(1,tt)-probe(ii)%vxmin)/probe(ii)%dvx) + 1
           rest(1)  = mod((vel(1,tt)-probe(ii)%vxmin),probe(ii)%dvx) 
           if ( Ng(1) < 1 ) then
              Ng(1) = 1
              rest(1) = 0.0d0
           end if
           if ( Ng(1) >= probe(ii)%Nvx ) then
              Ng(1) = probe(ii)%Nvx - 1
              rest(1) = probe(ii)%dvx
           end if
           drest(1) = probe(ii)%dvx-rest(1)

           Ng(2)    = int((vel(2,tt)-probe(ii)%vymin)/probe(ii)%dvy) + 1 
           rest(2)  = mod((vel(2,tt)-probe(ii)%vymin),probe(ii)%dvy) 
           if ( Ng(2) < 1 ) then
              Ng(2) = 1
              rest(2) = 0.0d0
           end if
           if ( Ng(2) >= probe(ii)%Nvy ) then
              Ng(2) = probe(ii)%Nvy - 1
              rest(2) = probe(ii)%dvy
           end if
           drest(2) = probe(ii)%dvy-rest(2)

           Ng(3)    = int((vel(3,tt)-probe(ii)%vzmin)/probe(ii)%dvz) + 1
           rest(3)  = mod((vel(3,tt)-probe(ii)%vzmin),probe(ii)%dvz) 
           if ( Ng(3) < 1 ) then
              Ng(3) = 1
              rest(3) = 0.0d0
           end if
           if ( Ng(3) >= probe(ii)%Nvz ) then
              Ng(3) = probe(ii)%Nvz - 1
              rest(3) = probe(ii)%dvz
           end if
           drest(3) = probe(ii)%dvz-rest(3)

           ! Start interpolation
           F1 = drest(1)*drest(2)*drest(3)/dVv
           F2 = rest(1)*drest(2)*drest(3)/dVv
           F3 = drest(1)*rest(2)*drest(3)/dVv
           F5 = rest(1)*rest(2)*drest(3)/dVv
           F4 = drest(1)*drest(2)*rest(3)/dVv
           F7 = rest(1)*drest(2)*rest(3)/dVv
           F6 = drest(1)*rest(2)*rest(3)/dVv
           F8 = rest(1)*rest(2)*rest(3)/dVv

           probe(ii)%f(Ng(1),Ng(2),Ng(3)) = &
                probe(ii)%f(Ng(1),Ng(2),Ng(3)) + lambda*F1
           probe(ii)%f(Ng(1)+1,Ng(2),Ng(3)) = &
                probe(ii)%f(Ng(1)+1,Ng(2),Ng(3)) + lambda*F2
           probe(ii)%f(Ng(1),Ng(2)+1,Ng(3)) = &
                probe(ii)%f(Ng(1),Ng(2)+1,Ng(3)) + lambda*F3
           probe(ii)%f(Ng(1)+1,Ng(2)+1,Ng(3)) = &
                probe(ii)%f(Ng(1)+1,Ng(2)+1,Ng(3)) + lambda*F5
           probe(ii)%f(Ng(1),Ng(2),Ng(3)+1) = &
                probe(ii)%f(Ng(1),Ng(2),Ng(3)+1) + lambda*F4
           probe(ii)%f(Ng(1)+1,Ng(2),Ng(3)+1) = &
                probe(ii)%f(Ng(1)+1,Ng(2),Ng(3)+1) + lambda*F7
           probe(ii)%f(Ng(1),Ng(2)+1,Ng(3)+1) = &
                probe(ii)%f(Ng(1),Ng(2)+1,Ng(3)+1) + lambda*F6
           probe(ii)%f(Ng(1)+1,Ng(2)+1,Ng(3)+1)= &
                probe(ii)%f(Ng(1)+1,Ng(2)+1,Ng(3)+1) + lambda*F8
        end if
     end do

  end do

end subroutine AccumulatefProbe

!------------------------------------------------------------------------
