! Subroutines used for various kinds of initialisation
!
!------------------------------------------------------------------------


subroutine InitialiseParticles_01(vel,pos,X,Y,Z,vx0,vy0,vz0,vth,Lside,Npart)

! This subroutine initialises a particle distribution in a cube the 
! side of which is Lside, and which is aligned with the axes of 
! the cartesian coordinate system.


! No implicit variables
  implicit none

  integer Npart, pp

  double precision X, Y, Z, vx0, vy0, vz0

  double precision vel(3,Npart),pos(3,Npart), vth,Lside,rndp(3)

  do pp = 1, Npart
     call random_number(rndp)
     pos(1,pp) = X + rndp(1)*Lside - 0.5d0*Lside      
     pos(2,pp) = Y + rndp(2)*Lside - 0.5d0*Lside      
     pos(3,pp) = Z + rndp(3)*Lside - 0.5d0*Lside      
  end do



! Use the subroutine Gaussian to produce thermal
! distributions.
! Maxwellian f(vx) ~ exp(-vx^2/vth^2) etc. => vth=sqrt(2*kB*T/m)

  call  Gaussian(vel(1,:),Npart,vx0,vth/sqrt(2.0d0))

  call  Gaussian(vel(2,:),Npart,vy0,vth/sqrt(2.0d0))

  call  Gaussian(vel(3,:),Npart,vz0,vth/sqrt(2.0d0))

  return
end subroutine InitialiseParticles_01

!------------------------------------------------------------------------

subroutine InitialiseParticles_02(vel, pos, Npart, mass, &
     scdatafilename, Ngridlevels, grid)

  use SpecificTypes

  implicit none

  integer Npart, Ngridlevels
  double precision B0(3), B2
  double precision mass, Eperp(3), E0(3)
  double precision vel(3,Npart), pos(3,Npart)
  character scdatafilename*200
  type (GridLevel) grid(Ngridlevels)
  
  !particles data read from spacecraft data file  
  double precision v_par(Npart), vth_perp(Npart)
  double precision indata(5)

  double precision unitB(3), unitE(3), vdrift(3)
  integer pp, over

  ! Initialise variables
  pos      = 0.0d0
  vel      = 0.0d0
  indata   = 0.0d0
  v_par    = 0.0d0
  vth_perp = 0.0d0
  over     = 0
  pp       = 0

! Read spacecraft data file
  open(unit=1,file=scdatafilename,status='old')
  do while ((over==0) .and. (pp<Npart))
     read(1,*, end=99, err=99) indata
     pp = pp + 1
     pos(:,pp)     = indata(1:3)
     v_par(pp)     = indata(4)
     vth_perp(pp)  = indata(5)
  end do
99 over = 1
  close(1)

  do pp = 1, Npart
     ! Find B field at the point where we are and call it B0
     ! Find E field at the point where we are and call it E0
     call FindFields(Ngridlevels, grid, pos(1,pp), pos(2,pp), pos(3,pp), E0, B0)

     ! squared magnitude of B0
     B2 = sum(B0**2.0d0)

     ! unit vector in the same direction as B
     unitB = B0/sqrt(B2)

     ! Compute perpendicular E, i.e. remove artificial parallel component
     call PerpendicularEfield(E0, B0, Eperp)

     ! unit vector in the same direction as Eperp
     unitE = Eperp / sqrt(sum(Eperp**2.0d0))

     ! Compute E cross B drift velocity
     vdrift(1) = ( Eperp(2)*B0(3) - Eperp(3)*B0(2) ) / B2
     vdrift(2) = ( Eperp(3)*B0(1) - Eperp(1)*B0(3) ) / B2
     vdrift(3) = ( Eperp(1)*B0(2) - Eperp(2)*B0(1) ) / B2

     vel(:,pp) = v_par(pp)*unitB + vdrift + vth_perp(pp)*unitE
  end do

  return

end subroutine InitialiseParticles_02

!------------------------------------------------------------------------

subroutine InitialiseParticles_03(vel, pos, &
     thetamin, thetamax, azmin, azmax, vr0, vth, rsource, Npart)

! This subroutine initialises a particle distribution on the surface 
! of a sphere with radius rsource. which has its centre at x=y=z=0.

! No implicit variables
  implicit none

  integer Npart
  double precision vel(3,Npart),pos(3,Npart)
  double precision thetamin, thetamax, azmin, azmax, vr0, vth, rsource

  integer pp
  double precision rndp(2), phi, theta, th, t
  double precision phimin, phimax, thmin, thmax, tmin, tmax, pi, vC(3)

  parameter(pi = 3.14159265358979323846264338327950288419716d0)

  ! convert to radians
  ! phi: azimuthal angle (angle to the positive y axis) in radians
  ! th:  theta (angle to positive x axis) converted to radians
  ! t:   variable in which the probability distribution is flat; t=-cos(th)
  phimin = azmin*pi/180.0d0
  phimax = azmax*pi/180.0d0
  thmin  = thetamin*pi/180.0d0
  thmax  = thetamax*pi/180.0d0
  tmin   = -cos(thmin)
  tmax   = -cos(thmax)

  do pp = 1, Npart
     call random_number(rndp)
     phi = phimin + rndp(1)*(phimax - phimin)
     t   = tmin  + rndp(2)*(tmax - tmin)
     th = pi - acos(t)
     pos(1,pp) = rsource*cos(th)
     pos(2,pp) = rsource*sin(th)*cos(phi)
     pos(3,pp) = rsource*sin(th)*sin(phi)
  end do



! Use the subroutines Gaussian and HalfMaxwellian to produce thermal
! distributions with positive radial components in a spherical 
! coordinate system
! Maxwellian f(vx) ~ exp(-vx^2/vth^2) etc. => vth=sqrt(2*kB*T/m)

  call HalfMaxwellian(vel(1,:),Npart,vr0,vth/sqrt(2.0d0))
  call Gaussian(vel(2,:),Npart,0.0d0,vth/sqrt(2.0d0))
  call Gaussian(vel(3,:),Npart,0.0d0,vth/sqrt(2.0d0))
 
! Transform to a Cartesian system.
  do pp = 1, Npart
     ! We use the phi and theta variables from above, although their 
     ! definitions have changed.
     theta = acos(pos(3,pp)/rsource)
     phi   = atan2(pos(2,pp), pos(1,pp))
     vC(1) = sin(theta)*cos(phi)*vel(1,pp) + cos(theta)*cos(phi)*vel(2,pp) &
          - sin(phi)*vel(3,pp)
     vC(2) = sin(theta)*sin(phi)*vel(1,pp) + cos(theta)*sin(phi)*vel(2,pp) &
          + cos(phi)*vel(3,pp)
     vC(3) = cos(theta)*vel(1,pp) - sin(theta)*vel(2,pp)
     ! Replace the spherical by the Cartesian velocity
     vel(:,pp) = vC
  end do

  return

end subroutine InitialiseParticles_03

!------------------------------------------------------------------------

subroutine Gaussian(x,N,mu,sigma)

! Generates a normally distributed sequence of random numbers
! f(v) ~ exp(-(v-mu)^2/(2*sigma^2)

! No implicit variables
  implicit none

! Parameters
  integer N
  double precision x(N), mu, sigma

! Local variables
  integer ii
  double precision pi, rndp(2)
  pi = 3.14159265358979323846264338327950288419716d0
  
  do ii = 1, N
     call random_number(rndp)
     x(ii) = sqrt(-2.0d0*log(rndp(1)))*cos(2.0d0*pi*rndp(2))*sigma + mu
  end do

  return 

end subroutine Gaussian

!------------------------------------------------------------------------

subroutine HalfMaxwellian(x,N,mu,sigma)

! Generates the positive half of a normal distribution. 
! f(v) ~ exp(-v^2/(2*sigma^2)
!
! If mu is not equal to 0.0d0 a shifted normal distribution 
! with a lower cutoff at 0.0d0 is generated.
! f(v) ~ H(v)*exp(-(v-mu)^2/(2*sigma^2)
! where H(v) = 0 for v<0 and 1 otherwise

  implicit none

! Parameters
  integer N
  double precision x(N), mu, sigma

! Local variables
  integer ii
  double precision pi, rndp(2)
  pi = 3.14159265358979323846264338327950288419716d0

  ! initialise
  x = -1.0d0

  if ( mu == 0.0d0 ) then
     ! Half Maxwellian
     do ii = 1, N
        call random_number(rndp)
        x(ii) = sqrt(-2.0d0*log(rndp(1)))*sin(pi*rndp(2))*sigma
     end do
  else
     ! Drifting Maxwellian with no negative velocities.
     do ii = 1, N
        do while ( x(ii) < 0.0d0 ) 
           call random_number(rndp)
           x(ii) = sqrt(-2.0d0*log(rndp(1)))*cos(2.0d0*pi*rndp(2))*sigma + mu
        end do
     end do
  end if

  return 

end subroutine HalfMaxwellian

!------------------------------------------------------------------------

recursive subroutine GetGeneralInput(inpar, attemptno, Nretries)
	   
  use SpecificTypes

  implicit none

  type (InputGeneral) inpar
  integer attemptno, Nretries

  character indata*132, filename*242
  integer i, j, k

  attemptno = attemptno + 1

! Defaults
  inpar%NhybSpecies= 1 ! Number of species in the hybrid simulation
  inpar%Xcube  = 0.0d0 ! x-coordinate [m] for centre of cube
  inpar%Ycube  = 0.0d0 ! y-coordinate [m] for centre of cube
  inpar%Zcube  = 1.0d6 ! z-coordinate [m] for centre of cube
  inpar%vxbulk = 0.0d0 ! x-component [m/s] of bulk velocity
  inpar%vybulk = 0.0d0 ! y-component [m/s] of bulk velocity
  inpar%vzbulk = 1.0d4 ! z-component [m/s] of bulk velocity
  inpar%rsource = 5.0d4 ! Radius of source sphere [m]
  inpar%vrbulk = 0.0d0 ! radial component [m/s] of bulk velocity
  inpar%Npart  = 1     ! Number of trajectories to compute
  inpar%Nsteps = 10000 ! Maximum number of time steps per trajectory
  inpar%Ngridlevels = 1 ! Number of grid levels
  inpar%vth    = 1.0d4 ! Thermal speed [m/s] for Maxwellian 
                       ! f(vx) ~ exp(-vx^2/vth^2) etc. => vth=sqrt(2*kB*T/m)
  inpar%Lside  = 1.0d4 ! Side of cube where particles start [m]
  inpar%dt     = 1.0d0 ! Time step [s]
  inpar%mass   = 2.67619d-26     ! Ion mass [kg]
  inpar%charge = 1.602176565d-19 ! Ion charge [As]
  inpar%robst = 1.0d4            ! Obstacle radius [m]
  inpar%initialiser = 1        ! Initialisation method
  inpar%scdatafilename='rosetta.dat' ! Input file name for initialiser=2
  inpar%thetamin = 0.0         ! Minimum angle to x axis for initialiser=3
  inpar%thetamax = 180.0       ! Maximum angle to x axis for initialiser=3
  inpar%azmin = 0.0            ! Minimum azimuth angle for initialiser=3
  inpar%azmax = 360.0          ! Maximum azimuth angle for initialiser=3
                               ! The y axis has theta=90 and azimuth=0, 
                               ! whereas the z axis has theta=90 and azimuth=90.
! Fluctuation model parameters
  inpar%Efluctuations = .false. ! Use fluctuation E-field or not
  inpar%rinfluct = 5.0d5   ! The field is applied between the cometocentric
  inpar%routfluct = 1.5d7  ! distances rinfluct and routfluct with a half
                           ! wavelength sinusoidal profile in between.
  inpar%E1x = 0.0d-4       ! Primary fluctuation field in the x-direction.
  inpar%E1y = 0.0d-4       ! Primary fluctuation field in the y-direction.
  inpar%E1z = 0.0d-4       ! Primary fluctuation field in the z-direction.
  inpar%ffluct = 1.0d-2;   ! Frequency of the fluctuations
  inpar%usealpha = .false. ! If 'yes' the field is adjusted to the specified
  inpar%alpha = 0.0d-2     ! angle [radians] between E and perpendicular
                           ! (to B) direction.

! Cheddar specific parameters
  inpar%Nprobes = 0                   ! Number of distribution function probes 
  inpar%startfromdumpfile = .false.
  inpar%dump_period_dump = 100000
  ! Output grid
  inpar%Nx = 60                  ! Number of grid points in x-direction 
  inpar%Ny = 60                  ! Number of grid points in y-direction 
  inpar%Nz = 60                  ! Number of grid points in z-direction 
  inpar%xmin = -17852810         ! Lower x-boundary [m]
  inpar%xmax =  17852810         ! Upper x-boundary [m]
  inpar%ymin = -17852810         ! Lower y-boundary [m]
  inpar%ymax =  17852810         ! Upper y-boundary [m]
  inpar%zmin = -17852810         ! Lower z-boundary [m]
  inpar%zmax =  17852810         ! Upper z-boundary [m]

! Photon pressure parameters
  inpar%includephotonpressure = .false. ! Whether to use Flight or not
  inpar%Flight = 0.0d-18       ! Force on dust particle from radiation pressure

! Common hybrid files
  inpar%magfilename='No file here'
  inpar%curfilename='No file here'
  
  filename='inputcheese.m'
  open(unit=1,file=filename,status='old',err=97)
  
  read (1,'(a)',err=98) indata

  do while (index(indata,'%END')+index(indata,'%end') == 0)
     i=index(indata,'=')
     if (i>0) then
        do j=1,i-1
           if(indata(j:j) /= ' ') exit
        enddo
        ! If there is a semicolon on the line, ignore it and everything beyond.
        k = index(indata(i+1:132),';')
        if (k==0) then
           k = 133-i
        end if
        select case (indata(j:i-1))
        case ('NhybSpecies')
           read (indata(i+1:i+k-1),*) inpar%NhybSpecies
        case ('Xcube')
           read (indata(i+1:i+k-1),*) inpar%Xcube
        case ('Ycube')
           read (indata(i+1:i+k-1),*) inpar%Ycube
        case ('Zcube')
           read (indata(i+1:i+k-1),*) inpar%Zcube
        case ('vxbulk')
           read (indata(i+1:i+k-1),*) inpar%vxbulk
        case ('vybulk')
           read (indata(i+1:i+k-1),*) inpar%vybulk
        case ('vzbulk')
           read (indata(i+1:i+k-1),*) inpar%vzbulk
        case ('rsource')
           read (indata(i+1:i+k-1),*) inpar%rsource
        case ('vrbulk')
           read (indata(i+1:i+k-1),*) inpar%vrbulk
        case ('Npart')
           read (indata(i+1:i+k-1),*) inpar%Npart
        case ('Nsteps')
           read (indata(i+1:i+k-1),*) inpar%Nsteps
        case ('Ngridlevels')
           read (indata(i+1:i+k-1),*) inpar%Ngridlevels
        case ('vth')
           read (indata(i+1:i+k-1),*) inpar%vth
        case ('Lside')
           read (indata(i+1:i+k-1),*) inpar%Lside
        case ('dt')
           read (indata(i+1:i+k-1),*) inpar%dt
        case ('mass')
           read (indata(i+1:i+k-1),*) inpar%mass
        case ('charge')
           read (indata(i+1:i+k-1),*) inpar%charge
        case ('robst')
           read (indata(i+1:i+k-1),*) inpar%robst
        case ('initialiser')
           read (indata(i+1:i+k-1),*) inpar%initialiser
        case ('scdatafilename')
           read (indata(i+1:i+k-1),*) inpar%scdatafilename
        case ('thetamin')
           read (indata(i+1:i+k-1),*) inpar%thetamin
        case ('thetamax')
           read (indata(i+1:i+k-1),*) inpar%thetamax
        case ('azmin')
           read (indata(i+1:i+k-1),*) inpar%azmin
        case ('azmax')
           read (indata(i+1:i+k-1),*) inpar%azmax
        case ('Efluctuations')
           if (index(indata(i+1:i+k-1),'''yes''') > 0) then
              inpar%Efluctuations = .true.
           end if
        case ('rinfluct')
           read (indata(i+1:i+k-1),*) inpar%rinfluct
        case ('routfluct')
           read (indata(i+1:i+k-1),*) inpar%routfluct
        case ('E1x')
           read (indata(i+1:i+k-1),*) inpar%E1x
        case ('E1y')
           read (indata(i+1:i+k-1),*) inpar%E1y
        case ('E1z')
           read (indata(i+1:i+k-1),*) inpar%E1z
        case ('ffluct')
           read (indata(i+1:i+k-1),*) inpar%ffluct
        case ('usealpha')
           if (index(indata(i+1:i+k-1),'''yes''') > 0) then
              inpar%usealpha = .true.
           end if
        case ('alpha')
           read (indata(i+1:i+k-1),*) inpar%alpha
        case ('Nx')
           read (indata(i+1:i+k-1),*) inpar%Nx
        case ('Ny')
           read (indata(i+1:i+k-1),*) inpar%Ny
        case ('Nz')
           read (indata(i+1:i+k-1),*) inpar%Nz
        case ('xmin')
           read (indata(i+1:i+k-1),*) inpar%xmin
        case ('xmax')
           read (indata(i+1:i+k-1),*) inpar%xmax
        case ('ymin')
           read (indata(i+1:i+k-1),*) inpar%ymin
        case ('ymax')
           read (indata(i+1:i+k-1),*) inpar%ymax
        case ('zmin')
           read (indata(i+1:i+k-1),*) inpar%zmin
        case ('zmax')
           read (indata(i+1:i+k-1),*) inpar%zmax
        case ('magfilename')
           read (indata(i+1:i+k-1),*) inpar%magfilename
        case ('curfilename')
           read (indata(i+1:i+k-1),*) inpar%curfilename
        case ('Nprobes')
           read (indata(i+1:i+k-1),*) inpar%Nprobes
        case ('startfromdumpfile')
           if (index(indata(i+1:i+k-1),'''yes''') > 0) then
              inpar%startfromdumpfile = .true.
           end if
        case ('dump_period_dump')
           read (indata(i+1:i+k-1),*) inpar%dump_period_dump
        case ('includephotonpressure')
           if (index(indata(i+1:i+k-1),'''yes''') > 0) then
              inpar%includephotonpressure = .true.
           end if
        case ('Flight')
           read (indata(i+1:i+k-1),*) inpar%Flight
        case ('saving_step_length')
           read (indata(i+1:i+k-1),*) inpar%saving_step_length
        case default
           if (index(indata(j:j+1),'%')==0) then
              write (*,*) 'Input quantity ', indata(j:i-1), ' is unknown.'
           end if
        end select
     end if
     read (1,'(a)',err=98) indata
  end do

  close(1,err=99)
  return

! Error handling section
97 write (*,*) 'GetGeneralInput: error in open statement'
  goto 100
98 write (*,*) 'GetGeneralInput: error in read statement'
  close(1,err=99)
  goto 100
99 write (*,*) 'GetGeneralInput: error in close statement'
100 if (attemptno<=Nretries) then
     call GetGeneralInput(inpar, attemptno, Nretries)
  end if

end subroutine GetGeneralInput

!------------------------------------------------------------------------

recursive subroutine GetSpecies(FromDir, Nspecies, particle, &
     attemptno, Nretries)

  use SpecificTypes

  implicit none

  character(len=*) :: FromDir
  integer Nspecies, attemptno, Nretries
  type(species) particle(Nspecies)

  character indata*132, filename*242
  integer ii, i, j, k

  attemptno = attemptno + 1

  filename=FromDir//'inputcheese.m'
  open(unit=1,file=filename,status='old',err=97)

  do ii = 1, Nspecies
     ! Defaults
     particle(ii)%hybfilename = 'No file name here, sorry.'
     particle(ii)%chargestate = 1.0d0
     particle(ii)%hybiteration = 0
     read (1,'(a)',err=98) indata
     do while (index(indata,'%SPEC')+index(indata,'%spec') == 0)
        read (1,'(a)') indata
     end do
     do while (index(indata,'%END')+index(indata,'%end') == 0)
        i=index(indata,'=')
        if (i>0) then
           do j=1,i-1
              if(indata(j:j) /= ' ') exit
           enddo
           ! If there is a semicolon on the line, ignore it and 
           ! everything beyond.
           k = index(indata(i+1:132),';')
           if (k==0) then
              k = 133-i
           end if
           select case (indata(j:i-1))
           case ('hybfilename')
              read (indata(i+1:i+k-1),*) particle(ii)%hybfilename
           case ('chargestate')
              read (indata(i+1:i+k-1),*) particle(ii)%chargestate
           case ('hybiteration')
              read (indata(i+1:i+k-1),*) particle(ii)%hybiteration
           case default
              if (index(indata(j:j+1),'%')==0) then
                 write (*,*) 'Input quantity ', indata(j:i-1), ' is unknown.'
              end if
           end select
        end if
        read (1,'(a)',err=98) indata
     end do
  end do

  close(1,err=99)
  return

! Error handling section
97 write (*,*) 'GetSpecies: error in open statement'
  goto 100
98 write (*,*) 'GetSpecies: error in read statement'
  close(1,err=99)
  goto 100
99 write (*,*) 'GetSpecies: error in close statement'
100 if (attemptno<=Nretries) then
     call GetSpecies(FromDir, Nspecies, particle, attemptno, Nretries)
  end if

end subroutine GetSpecies

!------------------------------------------------------------------------

recursive subroutine GetGridLevels(FromDir, Ngridlevels, grid, &
     attemptno, Nretries)

  use SpecificTypes

  implicit none

  character(len=*) :: FromDir
  integer Ngridlevels, attemptno, Nretries
  type(GridLevel) grid(Ngridlevels)

  character indata*132, filename*242
  integer ii, i, j, k

  attemptno = attemptno + 1

  filename=FromDir//'inputcheese.m'
  open(unit=1,file=filename,status='old',err=97)

  do ii = 1, Ngridlevels
     ! Defaults
     grid(ii)%Nx = 60                  ! Number of grid points in x-direction 
     grid(ii)%Ny = 60                  ! Number of grid points in y-direction 
     grid(ii)%Nz = 60                  ! Number of grid points in z-direction 
     grid(ii)%xmin = -17852810         ! Lower x-boundary [m]
     grid(ii)%xmax =  17852810         ! Upper x-boundary [m]
     grid(ii)%ymin = -17852810         ! Lower y-boundary [m]
     grid(ii)%ymax =  17852810         ! Upper y-boundary [m]
     grid(ii)%zmin = -17852810         ! Lower z-boundary [m]
     grid(ii)%zmax =  17852810         ! Upper z-boundary [m]

     read (1,'(a)',err=98) indata
     do while (index(indata,'%GRID')+index(indata,'%grid') == 0)
        read (1,'(a)') indata
     end do
     do while (index(indata,'%END')+index(indata,'%end') == 0)
        i=index(indata,'=')
        if (i>0) then
           do j=1,i-1
              if(indata(j:j) /= ' ') exit
           enddo
           ! If there is a semicolon on the line, ignore it and 
           ! everything beyond.
           k = index(indata(i+1:132),';')
           if (k==0) then
              k = 133-i
           end if
           select case (indata(j:i-1))
           case ('Nx')
              read (indata(i+1:i+k-1),*) grid(ii)%Nx
           case ('Ny')
              read (indata(i+1:i+k-1),*) grid(ii)%Ny
           case ('Nz')
              read (indata(i+1:i+k-1),*) grid(ii)%Nz
           case ('xmin')
              read (indata(i+1:i+k-1),*) grid(ii)%xmin
           case ('xmax')
              read (indata(i+1:i+k-1),*) grid(ii)%xmax
           case ('ymin')
              read (indata(i+1:i+k-1),*) grid(ii)%ymin
           case ('ymax')
              read (indata(i+1:i+k-1),*) grid(ii)%ymax
           case ('zmin')
              read (indata(i+1:i+k-1),*) grid(ii)%zmin
           case ('zmax')
              read (indata(i+1:i+k-1),*) grid(ii)%zmax
           case default
              if (index(indata(j:j+1),'%')==0) then
                 write (*,*) 'Input quantity ', indata(j:i-1), ' is unknown.'
              end if
           end select
        end if
        read (1,'(a)',err=98) indata
     end do
  end do
  close(1,err=99)
  return

! Error handling section
97 write (*,*) 'GetGridLevels: error in open statement'
  goto 100
98 write (*,*) 'GetGridLevels: error in read statement'
  close(1,err=99)
  goto 100
99 write (*,*) 'GetGridLevels: error in close statement'
100 if (attemptno<=Nretries) then
     call GetGridLevels(FromDir, Ngridlevels, grid, attemptno, Nretries)
  end if

end subroutine GetGridLevels

!------------------------------------------------------------------------

recursive subroutine GetProbes(FromDir, Nprobes, probe, &
     attemptno, Nretries)

  use SpecificTypes

  implicit none

  character(len=*) :: FromDir
  integer Nprobes, attemptno, Nretries
  type(fprobe) probe(Nprobes)

  character indata*132, filename*242
  integer ii, i, j, k

  attemptno = attemptno + 1

  filename=FromDir//'inputcheese.m'
  open(unit=1,file=filename,status='old',err=97)

  do ii = 1, Nprobes
     ! Defaults
     probe(ii)%Nvx = 10
     probe(ii)%Nvy = 10
     probe(ii)%Nvz = 10
     probe(ii)%vxmin = -1.0d6
     probe(ii)%vxmax =  1.0d6
     probe(ii)%vymin = -1.0d6
     probe(ii)%vymax =  1.0d6
     probe(ii)%vzmin = -1.0d6
     probe(ii)%vzmax =  1.0d6
     probe(ii)%xc = -5.0d6
     probe(ii)%yc =  0.0d0
     probe(ii)%zc =  1.0d7
     probe(ii)%rprobe = 5.0d5
     read (1,'(a)',err=98) indata
     do while (index(indata,'%PROB')+index(indata,'%prob') == 0)
        read (1,'(a)') indata
     end do
     do while (index(indata,'%END')+index(indata,'%end') == 0)
        i=index(indata,'=')
        if (i>0) then
           do j=1,i-1
              if(indata(j:j) /= ' ') exit
           enddo
           ! If there is a semicolon on the line, ignore it and 
           ! everything beyond.
           k = index(indata(i+1:132),';')
           if (k==0) then
              k = 133-i
           end if
           select case (indata(j:i-1))
           case ('Nvx')
              read (indata(i+1:i+k-1),*) probe(ii)%Nvx
           case ('Nvy')
              read (indata(i+1:i+k-1),*) probe(ii)%Nvy
           case ('Nvz')
              read (indata(i+1:i+k-1),*) probe(ii)%Nvz
           case ('vxmin')
              read (indata(i+1:i+k-1),*) probe(ii)%vxmin
           case ('vxmax')
              read (indata(i+1:i+k-1),*) probe(ii)%vxmax
           case ('vymin')
              read (indata(i+1:i+k-1),*) probe(ii)%vymin
           case ('vymax')
              read (indata(i+1:i+k-1),*) probe(ii)%vymax
           case ('vzmin')
              read (indata(i+1:i+k-1),*) probe(ii)%vzmin
           case ('vzmax')
              read (indata(i+1:i+k-1),*) probe(ii)%vzmax
           case ('xc')
              read (indata(i+1:i+k-1),*) probe(ii)%xc
           case ('yc')
              read (indata(i+1:i+k-1),*) probe(ii)%yc
           case ('zc')
              read (indata(i+1:i+k-1),*) probe(ii)%zc
           case ('rprobe')
              read (indata(i+1:i+k-1),*) probe(ii)%rprobe
           case default
              if (index(indata(j:j+1),'%')==0) then
                 write (*,*) 'Input quantity ', indata(j:i-1), ' is unknown.'
              end if
           end select
        end if
        read (1,'(a)',err=98) indata
     end do
  end do

  close(1,err=99)
  return

! Error handling section
97 write (*,*) 'GetProbes: error in open statement'
  goto 100
98 write (*,*) 'GetProbes: error in read statement'
  close(1,err=99)
  goto 100
99 write (*,*) 'GetProbes: error in close statement'
100 if (attemptno<=Nretries) then
     call GetProbes(FromDir, Nprobes, probe, attemptno, Nretries)
  end if

end subroutine GetProbes

!------------------------------------------------------------------------

recursive subroutine GetHybridData(FromDir, Nx, Ny, Nz, X, Y, Z, &
     dxyz, xmin, ymin, zmin, particle, attemptno, Nretries, ierr)

  use SpecificTypes

  implicit none

  character(len=*) :: FromDir
  integer Nx, Ny, Nz, attemptno, Nretries, ierr
  double precision  X(Nx), Y(Ny), Z(Nz), dxyz(3), xmin, ymin, zmin
  type(species) particle

  character indata*332, filename*242, variable(25)*20
  integer i0, ii, jj, kk, Nvar, over
  integer ix, iy, iz, irho, in, ivx, ivy, ivz, iBx, iBy, iBz, ijx, ijy, ijz
  integer xx, yy, zz
  double precision, allocatable :: inline(:)

  attemptno = attemptno + 1
  ierr = 0

  over = 0
  variable = ''
  ix=0; iy=0; iz=0; irho=0; in=0; ivx=0; ivy=0; ivz=0; 
  iBx=0; iBy=0; iBz=0; ijx=0; ijy=0; ijz=0

  filename=FromDir//'hybridfiles/'//particle%hybfilename
  open(unit=1,file=filename,status='old',err=97)
 
  ! read from the file until the column headings are found
  read (1,'(a)', err=98) indata
  do while (index(indata,'# X Y Z')+index(indata,'# x y z') == 0)
     read (1,'(a)', end=95, err=98) indata
     if ((index(indata,'1.602')>0) .and. (index(indata,'1.6021765300')==0)) then
        write (*,*) 'ERROR: Somebody has changed the elementary charge'
        stop
     end if
  end do
95 i0=index(indata,'#')
  ! read the column headings to find the variable names
  ii=i0
  if (i0>0) then
     do kk = 1, 25 ! twenty-five variables should be enough
        do jj = ii+1, 332
           if(indata(jj:jj) /= ' ') exit
        end do
        ii = jj + index(indata(jj+1:332),' ')
        if ((ii==jj) .or. (jj>332) .or. (ii>333)) exit
        read (indata(jj:ii-1),*) variable(kk)
        select case (variable(kk))
        case ('x')
           ix=kk
        case ('y')
           iy=kk
        case ('z')
           iz=kk
        case ('rho')
           irho=kk
        case ('n', 'Density')
           in=kk
        case ('vx', 'Ux')
           ivx=kk
        case ('vy', 'Uy')
           ivy=kk
        case ('vz', 'Uz')
           ivz=kk
        case ('Bx')
           iBx=kk
        case ('By')
           iBy=kk
        case ('Bz')
           iBz=kk
        case ('jx', 'Jx')
           ijx=kk
        case ('jy', 'Jy')
           ijy=kk
        case ('jz', 'Jz')
           ijz=kk
        end select
     end do
  else
     write (*,*) 'GetHybridData: No column header found in file ', filename
     goto 96
  end if
  Nvar = kk-1 ! Number of variables in the file
  allocate(inline(Nvar))
  ! read the file, line by line, and put the data in the right place
  do while (over==0)
     read(1,*, end=96, err=98) inline
     ! Find the correct grid point
     xx = nint( (inline(ix)-xmin)/dxyz(1))+1
     yy = nint( (inline(iy)-ymin)/dxyz(2))+1
     zz = nint( (inline(iz)-zmin)/dxyz(3))+1
     ! Insert the data
     if (iBx>0) then
        particle%B(1,xx,yy,zz) = inline(iBx)
     end if
     if (iBy>0) then
        particle%B(2,xx,yy,zz) = inline(iBy)
     end if
     if (iBz>0) then
        particle%B(3,xx,yy,zz) = inline(iBz)
     end if
     if (ivx>0) then
        particle%vel(1,xx,yy,zz) = inline(ivx)
     end if
     if (ivy>0) then
        particle%vel(2,xx,yy,zz) = inline(ivy)
     end if
     if (ivz>0) then
        particle%vel(3,xx,yy,zz) = inline(ivz)
     end if
     if (ijx>0) then
        particle%cur(1,xx,yy,zz) = inline(ijx)
     end if
     if (ijy>0) then
        particle%cur(2,xx,yy,zz) = inline(ijy)
     end if
     if (ijz>0) then
        particle%cur(3,xx,yy,zz) = inline(ijz)
     end if
     if (irho>0) then
        particle%rho(xx,yy,zz) = inline(irho)
     end if
     if (in>0) then
        particle%dens(xx,yy,zz) = inline(in)
     end if
  end do
96 over = 1
  close(1,err=99)
  return

! Error handling section
97 ierr = 97 
  write (*,*) 'GetHybridData: error in open statement, file: ', filename
  goto 100
98 ierr = 98 
  write (*,*) 'GetHybridData: error in read statement, file: ', filename
  close(1,err=99)
  goto 100
99 ierr = 99
  write (*,*) 'GetHybridData: error in close statement, file: ', filename
100 if (attemptno<=Nretries) then
     call GetHybridData(FromDir, Nx, Ny, Nz, X, Y, Z, &
          dxyz, xmin, ymin, zmin, particle, attemptno, Nretries, ierr)
  end if

end subroutine GetHybridData

!------------------------------------------------------------------------
