#!/usr/bin/php
<?php

// The sky's the limit. Note that when running PHP from the command line, there is no limit
// on the execution time.
//ini_set('memory_limit', '1G');

// Maximum number of points that can be handled at once.

define('MAX_DATA_POINTS', 50000);

// List of variables available for each resource ID.

$GLOBALS['VARIABLES'] = array(
    'He++_ave_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'H2O+_ave_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'H+_ave_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'He++_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'H2O+_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'H+_hybstate' => array('Density', 'Ux', 'Uy', 'Uz', 'Utot', 'Pressure', 'Temperature'),
    'Mag' => array('Bx', 'By', 'Bz', 'Btot'),
    'ElectricCurrent' => array('Jx', 'Jy', 'Jz', 'Jtot')
);

// Check and read command line arguments.

if ($argc != 5) {
    echo "Usage: $argv[0] MODEL NUMERICAL_OUTPUT INPUT_FILE OUTPUT_FILE\n";
    echo "\nExample: $argv[0] comet_1.3AU_20140315 H2O+_hybstate points.dat results.dat\n";
    exit(1);
}

$model = $argv[1];
$numericalOutput = $argv[2];
$inputFile = $argv[3];
$outputFile = $argv[4];

// Open input file for reading.

if (($input = fopen($inputFile, 'r')) === false) {
    exit(1);
}

// Open the output file for writing and write header line.

if (($output = fopen($outputFile, 'w')) === false) {
    exit(1);
}

fprintf($output, "# x y z %s\n", implode(' ', $GLOBALS['VARIABLES'][$numericalOutput]));

// Read input file into z, y, z arrays.

$xCoordinates = array();
$yCoordinates = array();
$zCoordinates = array();

while ($point = fscanf($input, "%f %f %f")) {
    $xCoordinates[] = $point[0];
    $yCoordinates[] = $point[1];
    $zCoordinates[] = $point[2];

    // Split input data into smaller datasets to handle server limitation.
    if (count($xCoordinates) == MAX_DATA_POINTS) {
        getDataPointValue($model, $numericalOutput, $xCoordinates, $yCoordinates, $zCoordinates, $output);

        // Reset arrays for the next dataset.
        unset($xCoordinates);
        $xCoordinates = array();
        unset($yCoordinates);
        $yCoordinates = array();
        unset($zCoordinates);
        $zCoordinates = array();
    }
}

// Process last dataset.
if (!empty($xCoordinates)) {
    getDataPointValue($model, $numericalOutput, $xCoordinates, $yCoordinates, $zCoordinates, $output);
}

// Close files and exit.

fclose($input);
fclose($output);
exit(0);

function getDataPointValue($model, $numericalOutput, &$xCoordinates, &$yCoordinates, &$zCoordinates, $handle)
{
    // Create SOAP client for communicating with FMI web service.
    
    $wsdlFile = 'http://impex-fp7.fmi.fi/ws/Methods_FMI.wsdl';
    $client = new SoapClient($wsdlFile);

    // Create the VOTable file that will contains the data points by calling the
    // corresponding method of the web service.

    $params = array(
        'Table_name' => 'Cheese Coordinates',
        'Description' => 'Cheese Description',
        'Fields' => array(
            array( 'name' => 'X', 'data' => $xCoordinates),
            array( 'name' => 'Y', 'data' => $yCoordinates),
            array( 'name' => 'Z', 'data' => $zCoordinates)
        )
    );

    $xyzURL = $client->getVOTableURL($params);

    // Compute the values of the physical quantities for the given points. This
    // returns the URL of the VOTable file containing the results.

    $params = array (
        'ResourceID' => "impex://FMI/HWA/HYB/comet/$model/$numericalOutput",
        'url_XYZ' => $xyzURL,
        'Variable' => implode(',', $GLOBALS['VARIABLES'][$numericalOutput]),
        'extraParams' => array('OutputFileType' => 'VOTable')
    );

    $dataURL = $client->getDataPointValue($params);
    echo "$dataURL\n";

    // Open the VOTable file in a XML Reader and open the output file for writing.

    $xml = new XMLReader();
    $xml->open($dataURL);

    // Find the beginning of tabular data.

    while ($xml->read() && !openingElement($xml, 'tabledata'));

    // Until the end of tabular data, read the rows. Note that the parser also stops
    // on text nodes that contain whitespaces. Those nodes must be skipped.

    for ($i = 1; $xml->read() && !closingElement($xml, 'tabledata'); ++$i) {
        // Find the beginning of a row (<tr>).
        while ($xml->read() && !openingElement($xml, 'tr'));

        // Read all the columns (<td>) and write them in the output file.
        while ($xml->read() && !closingElement($xml, 'tr')) {
            if (openingElement($xml, 'td')) {
                fprintf($handle, "%s ", $xml->readInnerXML());
            }
        }

        // Write EOL in output file only if a row was found.
        if (closingElement($xml, 'tr')) {
            fprintf($handle, "\n");
        }
    }

    // Close XML reader.

    $xml->close();
}

function openingElement($element, $name) {
    return !strcasecmp($element->name, $name) && $element->nodeType == XMLReader::ELEMENT;
}

function closingElement($element, $name) {
    return !strcasecmp($element->name, $name) && $element->nodeType == XMLReader::END_ELEMENT;
}
